import truss_chains as chains


class HelloWorld(chains.ModelBase):
    def __init__(self, context: chains.DeploymentContext = chains.depends_context()):
        self._call_count = 0

    def predict(self, call_count_increment: int) -> int:
        self._call_count += call_count_increment
        return self._call_count



import contextlib
import enum
import inspect
import logging
import os
import random
from typing import Any, Iterable, Iterator, TypeVar, Union

import pydantic
import pydantic_core

from truss_chains import public_types

T = TypeVar("T")


def make_abs_path_here(file_path: str) -> "public_types.AbsPath":
    """Helper to specify file paths relative to the *immediately calling* module.

    E.g. in you have a project structure like this::

        root/
            chain.py
            common_requirements.text
            sub_package/
                chainlet.py
                chainlet_requirements.txt

    You can now in ``root/sub_package/chainlet.py`` point to the requirements
    file like this::

        shared = make_abs_path_here("../common_requirements.text")
        specific = make_abs_path_here("chainlet_requirements.text")


    Warning:
        This helper uses the directory of the immediately calling module as an
        absolute reference point for resolving the file location. Therefore,
        you MUST NOT wrap the instantiation of ``make_abs_path_here`` into a
        function (e.g. applying decorators) or use dynamic code execution.

        Ok::

            def foo(path: AbsPath):
                abs_path = path.abs_path


            foo(make_abs_path_here("./somewhere"))

        Not Ok::

            def foo(path: str):
                dangerous_value = make_abs_path_here(path).abs_path


            foo("./somewhere")

    """
    # TODO: the absolute path resolution below uses the calling module as a
    #   reference point. This would not work if users wrap this call in a function
    #   - we hope the naming makes clear that this should not be done.
    caller_frame = inspect.stack()[1]
    module_path = caller_frame.filename
    if not os.path.isabs(file_path):
        module_dir = os.path.dirname(os.path.abspath(module_path))
        abs_file_path = os.path.normpath(os.path.join(module_dir, file_path))
    else:
        abs_file_path = file_path

    return public_types.AbsPath(abs_file_path, module_path, file_path)


def setup_dev_logging(level: Union[int, str] = logging.INFO) -> None:
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    log_format = "%(levelname)s %(asctime)s %(filename)s:%(lineno)d] %(message)s"
    date_format = "%m%d %H:%M:%S"
    formatter = logging.Formatter(fmt=log_format, datefmt=date_format)
    if root_logger.handlers:
        for handler in root_logger.handlers:
            handler.setFormatter(formatter)
    else:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)


@contextlib.contextmanager
def log_level(level: int) -> Iterator[None]:
    """Change loglevel for code in this context."""
    current_logging_level = logging.getLogger().getEffectiveLevel()
    logging.getLogger().setLevel(level)
    try:
        yield
    finally:
        logging.getLogger().setLevel(current_logging_level)


def expect_one(it: Iterable[T]) -> T:
    """Assert that an iterable has exactly on element and return it."""
    it = iter(it)
    try:
        element = next(it)
    except StopIteration:
        raise ValueError("Iterable is empty.")

    try:
        _ = next(it)
    except StopIteration:
        return element

    raise ValueError("Iterable has more than one element.")


########################################################################################


class InjectedError(Exception):
    """Test error for debugging/dev."""


def random_fail(probability: float, msg: str):
    """Probabilistically raises `InjectedError` for debugging/dev."""
    if random.random() < probability:
        print(f"Random failure: {msg}")
        raise InjectedError(msg)


class StrEnum(str, enum.Enum):
    """
    Adapted from MIT-licensed
    https://github.com/irgeek/StrEnum/blob/master/strenum/__init__.py

    This is useful for Pydantic-based (de-)serialisation, as Pydantic takes the value
    of an enum member as the value to be (de-)serialised, and not the name of the
    member. With this, we can have the member name and value be the same by using
    `enum.auto()`.

    StrEnum is a Python `enum.Enum` that inherits from `str`. The `auto()` behavior
    uses the member name and lowers it. This is useful for compatibility with pydantic.
    Example usage:

    ```
    class Example(StrEnum):
        SOME_VALUE = enum.auto()
        ANOTHER_VALUE = enum.auto()
        TEST = enum.auto()

    assert Example.SOME_VALUE == "SOME_VALUE"
    assert Example.ANOTHER_VALUE.value == "ANOTHER_VALUE"
    assert Example.TEST.value == Example.TEST
    assert Example.TEST == Example("TEST")
    ```
    """

    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, str):
            raise TypeError(f"Values of StrEnums must be strings: Got `{repr(value)}`.")
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self) -> str:
        return str(self.value)

    def _generate_next_value_(name, *_) -> str:  # type: ignore[override]
        if name.upper() != name:
            raise ValueError(f"Python enum members should be upper case. Got `{name}`.")
        return name


def issubclass_safe(x: Any, cls: type) -> bool:
    """Like built-in `issubclass`, but works on non-type objects."""
    return isinstance(x, type) and issubclass(x, cls)


def get_pydantic_field_default_value(
    model: type[pydantic.BaseModel], field_name: str
) -> Any:
    """Retrieve the default value of a field, considering both default and default_factory."""
    field_info = model.model_fields[field_name]
    if field_info.default is not pydantic_core.PydanticUndefined:
        return field_info.default
    if field_info.default_factory is not None:
        return field_info.default_factory()  #  type: ignore[call-arg]
    return None


def make_optional_import_error(module_name: str) -> public_types.ChainsRuntimeError:
    return public_types.ChainsRuntimeError(
        f"Could not import `{module_name}`. For chains CLI (truss package) this is an "
        "optional dependency. In deployed chainlets this dependency is "
        "automatically added. If you happen to run into this error, "
        f"install `{module_name}` manually."
    )



import importlib.metadata
import pathlib

import tomlkit


def _get_version() -> str:
    # For in-repo, `pyproject` takes precedence, but we need to assert it's
    # matching the truss package, because some other packages pollute `site-packages`
    # for pypi installs.
    toml_file = pathlib.Path(__file__).parent.parent / "pyproject.toml"
    if toml_file.exists():
        try:
            pyproject = tomlkit.parse(toml_file.read_text())
            poetry_section = pyproject["tool"]["poetry"]  # type: ignore[index]
            if poetry_section["name"] == __name__:  # type: ignore[index]
                return str(poetry_section["version"]).strip()  # type: ignore[index]
        except Exception:
            pass
    # Either there is no pyproject file or it's from a different package.
    # Try dist info metadata. This must be present.
    return importlib.metadata.version(__name__).strip()


__version__ = _get_version()


from truss.api import login, push, whoami
from truss.base import truss_config
from truss.truss_handle.build import load  # TODO: Refactor all usages and remove.

__all__ = ["push", "login", "load", "whoami", "truss_config"]



import importlib
import os
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import opentelemetry.sdk.trace as sdk_trace
import pytest
import yaml
from starlette.requests import Request


@pytest.fixture
def anyio_backend():
    return "asyncio"


@pytest.fixture
def connected_request():
    mock_request = MagicMock(spec=Request)
    mock_request.is_disconnected = AsyncMock(return_value=False)
    return mock_request


@pytest.fixture
def app_path(truss_container_fs: Path, helpers: Any):
    truss_container_app_path = truss_container_fs / "app"
    model_file_content = """
class Model:
    def __init__(self):
        self.load_count = 0

    def load(self):
        self.load_count += 1
        if self.load_count <= 2:
            raise RuntimeError('Simulated error')

    def predict(self, request):
        return request
    """
    with (
        helpers.file_content(
            truss_container_app_path / "model" / "model.py", model_file_content
        ),
        helpers.sys_path(truss_container_app_path),
    ):
        yield truss_container_app_path


@pytest.mark.anyio
async def test_model_wrapper_load_error_once(app_path, connected_request):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    os.chdir(app_path)
    model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
    model_wrapper.load()
    # Allow load thread to execute
    time.sleep(1)
    output = await model_wrapper.predict({}, connected_request)
    assert output == {}
    assert model_wrapper._model.load_count == 2


def test_model_wrapper_load_error_more_than_allowed(app_path, helpers):
    with helpers.env_var("NUM_LOAD_RETRIES_TRUSS", "0"):
        if "model_wrapper" in sys.modules:
            model_wrapper_module = sys.modules["model_wrapper"]
            importlib.reload(model_wrapper_module)
        else:
            model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())
        os.chdir(app_path)
        model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
        model_wrapper.load()
        # Allow load thread to execute
        time.sleep(1)
        assert model_wrapper.load_failed


@pytest.mark.anyio
@pytest.mark.integration
async def test_model_wrapper_streaming_timeout(app_path):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")

    # Create an instance of ModelWrapper with streaming_read_timeout set to 5 seconds
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    config["runtime"]["streaming_read_timeout"] = 5
    model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
    model_wrapper.load()
    assert model_wrapper._config.get("runtime").get("streaming_read_timeout") == 5


@pytest.mark.anyio
async def test_trt_llm_truss_predict(
    trt_llm_truss_container_fs, helpers, connected_request
):
    app_path = trt_llm_truss_container_fs / "app"
    packages_path = trt_llm_truss_container_fs / "packages"
    with (
        _clear_model_load_modules(),
        helpers.sys_paths(app_path, packages_path),
        _change_directory(app_path),
    ):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())

        expected_predict_response = "test"
        mock_predict_called = False

        async def mock_predict(return_value, request):
            nonlocal mock_predict_called
            mock_predict_called = True
            return expected_predict_response

        mock_engine = Mock(predict=mock_predict)
        mock_extension = Mock()
        mock_extension.load = Mock()
        mock_extension.model_args = Mock(return_value={"engine": mock_engine})
        with patch.object(
            model_wrapper_module, "_init_extension", return_value=mock_extension
        ):
            model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
            model_wrapper.load()
            resp = await model_wrapper.predict({}, connected_request)
            mock_extension.load.assert_called()
            mock_extension.model_args.assert_called()
            assert mock_predict_called
            assert resp == expected_predict_response


@pytest.mark.asyncio
async def test_trt_llm_truss_missing_model_py(
    trt_llm_truss_container_fs, helpers, connected_request
):
    app_path = trt_llm_truss_container_fs / "app"
    (app_path / "model" / "model.py").unlink()

    packages_path = trt_llm_truss_container_fs / "packages"
    with (
        _clear_model_load_modules(),
        helpers.sys_paths(app_path, packages_path),
        _change_directory(app_path),
    ):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())

        expected_predict_response = "test"
        mock_predict_called = False

        async def mock_predict(return_value, request: Request):
            nonlocal mock_predict_called
            mock_predict_called = True
            return expected_predict_response

        mock_engine = Mock(predict=mock_predict, spec=["predict"])
        mock_extension = Mock()
        mock_extension.load = Mock()
        mock_extension.model_override = Mock(return_value=mock_engine)
        with patch.object(
            model_wrapper_module, "_init_extension", return_value=mock_extension
        ):
            model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
            model_wrapper.load()
            resp = await model_wrapper.predict({}, connected_request)
            mock_extension.load.assert_called()
            mock_extension.model_override.assert_called()
            assert mock_predict_called
            assert resp == expected_predict_response


@pytest.mark.anyio
async def test_open_ai_completion_endpoints(
    open_ai_container_fs, helpers, connected_request
):
    app_path = open_ai_container_fs / "app"
    with (
        _clear_model_load_modules(),
        helpers.sys_paths(app_path),
        _change_directory(app_path),
    ):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())

        model_wrapper = model_wrapper_class(config, sdk_trace.NoOpTracer())
        model_wrapper.load()

        predict_resp = await model_wrapper.predict({}, connected_request)
        assert predict_resp == "predict"

        completions_resp = await model_wrapper.completions({}, connected_request)
        assert completions_resp == "completions"

        chat_completions_resp = await model_wrapper.chat_completions(
            {}, connected_request
        )
        assert chat_completions_resp == "chat_completions"


@contextmanager
def _change_directory(new_directory: Path):
    original_directory = os.getcwd()
    os.chdir(str(new_directory))
    try:
        yield
    finally:
        os.chdir(original_directory)


@contextmanager
def _clear_model_load_modules():
    """Clear dangling references to model and model_wrapper modules

    We do this before to clear any debris from before, and after to clean up
    after self. This is meant for cases where we simulate running a truss model
    in process, where these modules are loaded dyamically.
    """
    # These are left over by TrussModuleLoader used by local prediction tests.
    # TODO(pankaj) Find a way for TrussModuleLoader to clean up after itself.
    _remove_model_load_sys_modules()
    yield
    _remove_model_load_sys_modules()


def _remove_model_load_sys_modules():
    if "model" in sys.modules:
        del sys.modules["model"]
    if "model.model" in sys.modules:
        del sys.modules["model.model"]
    if "model_wrapper" in sys.modules:
        del sys.modules["model_wrapper"]



from truss.truss_handle.patch.dir_signature import directory_content_signature


def test_directory_content_signature(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "file1").touch()
    (root / "file2").touch()
    subdir = root / "dir"
    subdir.mkdir()
    (subdir / "file3").touch()

    content_sign = directory_content_signature(root)

    assert content_sign.keys() == {"dir", "dir/file3", "file1", "file2"}


def test_directory_content_signature_ignore_patterns(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "file1").touch()
    (root / "file2").touch()
    data_dir = root / "data"
    data_dir.mkdir()
    (data_dir / "file3").touch()

    git_dir = root / ".git"
    git_dir.mkdir()
    (git_dir / "HEAD").touch()
    git_subdir = git_dir / "objects"
    git_subdir.mkdir()
    (git_subdir / "00").touch()

    content_sign = directory_content_signature(
        root=root, ignore_patterns=["data/*", ".git"]
    )

    assert content_sign.keys() == {"data", "file1", "file2"}
